import os
import json
from tunas2dsdl.general.utils import Util

try:
    import ruamel_yaml as yaml
    from ruamel_yaml import YAML
except ImportError:
    import ruamel.yaml as yaml
    from ruamel.yaml import YAML


class Generation:
    LINESEP_NUM = 1

    def __init__(self, dsdl_version, meta_info, def_info, class_info, data_info, working_dir, separate_flag):
        self._working_dir = working_dir
        self._dsdl_version = dsdl_version
        self._meta_info = meta_info
        self._def_info = def_info
        self._data_info = data_info
        self._class_info = class_info
        self._yaml_obj = self._prepare_yaml_obj()
        self.separate_flag = separate_flag

    def _prepare_yaml_obj(self):
        default_yaml = YAML()
        class_dom_yaml = YAML()
        class_dom_yaml.indent(mapping=4, sequence=4, offset=4)
        sample_yaml = YAML()
        sample_yaml.indent(mapping=4, sequence=4, offset=2)

        result = {"class_dom": class_dom_yaml, "dsdl_version": default_yaml, "meta_info": default_yaml,
                  "struct_def": class_dom_yaml, "samples": sample_yaml, "import": class_dom_yaml}
        return result

    def write_meta(self, file_name=None, path=None):
        meta_info = self._meta_info.copy()
        for k, v in meta_info.items():
            meta_info[k] = Util.add_quotes(v)
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"
        with open(path, 'a') as fp:
            self._yaml_obj["meta_info"].dump({"meta": meta_info}, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)

    def write_dsdl_version(self, file_name=None, path=None):
        dsdl_version_info = self._dsdl_version.copy()
        for k, v in dsdl_version_info.items():
            dsdl_version_info[k] = Util.add_quotes(v)
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"
        with open(path, 'a') as fp:
            self._yaml_obj["dsdl_version"].dump(dsdl_version_info, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)

    def write_class_dom(self, file_name=None):
        file_name = file_name if file_name else self._class_info[-1].name
        dst = os.path.join(self._working_dir, file_name + ".yaml")
        self.write_dsdl_version(path=dst)
        with open(dst, "a") as fp:
            for class_domain in self._class_info:
                if not class_domain:
                    continue
                self._yaml_obj["class_dom"].dump(class_domain.format(), fp)
                fp.writelines([os.linesep] * self.LINESEP_NUM)
        return file_name

    def _write_struct_item(self, struct_item, file_name=None, path=None):
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"

        with open(path, "a") as fp:
            self._yaml_obj["struct_def"].dump(struct_item.format(), fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)

    def write_struct_def(self, file_name):

        dst = os.path.join(self._working_dir, file_name + ".yaml")
        self.write_dsdl_version(path=dst)

        for struct_def_item in self._def_info:
            self._write_struct_item(struct_def_item, file_name)
        return file_name

    def write_import_list(self, import_list=None, file_name=None, path=None):
        if not import_list:
            return
        if file_name is not None:
            path = os.path.join(self._working_dir, file_name + ".yaml")
        else:
            assert path is not None, "you must provide `file_name` or `path` args"
        with open(path, "a") as fp:
            self._yaml_obj["import"].dump({"$import": import_list}, fp)
            fp.writelines([os.linesep] * self.LINESEP_NUM)
        return file_name

    def write_samples(self, file_name, import_list=None):

        yaml_file = file_name + ".yaml"
        dst = os.path.join(self._working_dir, yaml_file)
        self.write_dsdl_version(path=dst)
        self.write_import_list(import_list, path=dst)
        self.write_meta(path=dst)

        if not self.separate_flag:
            with open(dst, "a") as fp:
                self._yaml_obj["samples"].dump(self._data_info.format(), fp)
        else:
            sample_file = "samples.json"
            data_info = self._data_info.format(sample_file)
            _data = data_info['data']
            samples = _data.pop("samples")
            samples = {"samples": samples}
            with open(dst, "a") as fp:
                self._yaml_obj["samples"].dump(data_info, fp)
            with open(dst.replace(yaml_file, sample_file), 'w') as fp:
                json.dump(samples, fp)



from tunas2dsdl.generator.generate import Generation
from tunas2dsdl.parser import DetectionParse
from .general.classdomain import ClassDomain, Label
from .general.struct import Struct, Field
from .general import OptionEatAll

__all__ = [
    "Generation",
    "DetectionParse",
    "ClassDomain",
    "Label",
    "Struct",
    "Field",
    "OptionEatAll",
]



from tunas2dsdl import DetectionParse, Generation
import os
import click
from tunas2dsdl import OptionEatAll
from tunas2dsdl.general import prepare_input


@click.command(name='convert')
@click.option("-d", "--directory", "directory", type=str,
              help="The directory path of the whole tunas v0.3 dataset.")
@click.option("-i", "--dataset-info", "dataset_info", type=str,
              help="The path of the dataset_info.json in tunas format dataset.")
@click.option("-a", "--annotation", "annotation_info", cls=OptionEatAll, type=str,
              help="The path of the annotation json file in tunas format dataset.")
@click.option("-w", "--wording-dir", "working_dir", type=str, required=True,
              help="The working directory where you want to generate the dsdl yaml files.")
@click.option("-t", "--task", "task", type=click.Choice(["detection", "classification"]), required=True,
              help="The task type you are working on. 'detection' and 'classification' are available.")
@click.option("-c", "--config", "config", type=str, required=False,
              help="The configuration json file of aliyun oss")
@click.option("-s", "--separate-store", "separate_store_flag", is_flag=True,
              help="Whether to store the samples in json file")
@prepare_input
def convert(directory, dataset_info, annotation_info, working_dir, task, config, separate_store_flag):
    if task == "detection":
        for seg_name, ann_file in annotation_info.items():
            conversion = DetectionParse(dataset_info, ann_file, separate_flag=separate_store_flag, file_client=config)
            this_work_dir = os.path.join(working_dir, seg_name)
            assert not os.path.isdir(this_work_dir), f"'{this_work_dir}' have existed, which is not permitted."
            os.mkdir(this_work_dir)
            generate_obj = Generation(conversion.dsdl_version, conversion.meta_info, conversion.struct_defs,
                                      conversion.class_domain_info, conversion.samples, this_work_dir,
                                      separate_flag=separate_store_flag)
            class_file = generate_obj.write_class_dom()
            def_file = generate_obj.write_struct_def(file_name=f"{seg_name}_object-detection")
            generate_obj.write_samples(file_name=seg_name, import_list=[class_file, def_file])
    elif task == "classification":
        pass


@click.group()
def cli():
    pass


cli.add_command(convert)

if __name__ == '__main__':
    cli()



import json
from ..general import ClassDomain, Label, Field, Struct, DataSample, Util
from tqdm import tqdm
from ..io import HardDiskBackend, FileIO


class DetectionParse:
    INSTANCE_SEG_ANN_TYPE = "polygon"
    DETECTION_ANN_TYPE = "box2d"
    PARENT_PATTERN = "Parent"

    def __init__(self, data_info_file, ann_file, separate_flag, dataset_name=None, file_client=None):
        if file_client:
            self.file_client = file_client
        else:
            self.file_client = HardDiskBackend()
        self.separate_flag = separate_flag
        self._dataset_name = dataset_name
        self._data_info_file = data_info_file
        self._annotation_file = ann_file
        self._data_info = self._read_json(data_info_file)
        self._annotation_info = self._read_json(ann_file)
        self._meta_info = self._parse_meta_info()
        self._parent_class_domain, self._class_domain, self.raw_class_info = self._parse_class_domain()  # 父类的class domain和子类的class domain
        self._dsdl_version_info = {"$dsdl-version": "0.5.0"}
        self._struct_defs, self._samples = self._parse_ann_info()

    def _parse_meta_info(self):
        """
        提取数据集元信息
        :return: 元信息字典
        """
        meta_info = {k: v for k, v in self._data_info.items() if k not in ("tasks", "statistics")}
        meta_info["sub_dataset_name"] = self._annotation_info["sub_dataset_name"]
        if not self._dataset_name:
            self._dataset_name = meta_info["dataset_name"]
        return meta_info

    def _parse_class_domain(self):
        """
        提取class domain相关的信息
        :return: 父类class domain对象，子类class domain对象，class原始信息字典
        """
        class_domain = ClassDomain(FileIO.clean(f"{self._dataset_name}ClassDom"))
        parent_class_domain = ClassDomain(FileIO.clean(f"{self._dataset_name}{self.PARENT_PATTERN}ClassDom"))
        raw_class_info = {}
        for task in tqdm(self._data_info["tasks"], desc="parsing class domain..."):
            task_type, catalog = task["type"], task["catalog"]
            task_class_info = {}
            if task_type not in (self.INSTANCE_SEG_ANN_TYPE, self.DETECTION_ANN_TYPE):
                continue
            raw_class_info[task_type] = task_class_info
            for label_info in catalog:
                label_name = label_info["category_name"]
                task_class_info[label_info['category_id']] = label_name
                label_obj = Label(label_name)
                if label_obj in class_domain:
                    continue
                class_domain.add_label(label_obj)
                for supercategory_name in label_info.get("supercategories", []):
                    super_label = Label(supercategory_name)
                    if super_label not in parent_class_domain:
                        parent_class_domain.add_label(super_label)
                    label_obj.add_supercategory(super_label)  # 添加父类
        if parent_class_domain:
            class_domain.set_parent(parent_class_domain)
            return parent_class_domain, class_domain, raw_class_info
        else:
            return None, class_domain, raw_class_info

    def _parse_ann_info(self):
        samples = self._annotation_info["samples"]
        object_struct = Struct("LocalObjectEntry")
        sample_struct = Struct("ObjectDetectionSample")
        media_field = Field("media", field_type="media")
        sample_struct.add_field(media_field)  # 添加图像field
        sample_container = DataSample(sample_struct, self._class_domain)
        annotation_option_flag = False
        for sample in tqdm(samples, desc="parsing samples ...."):
            sample_item = {}  # 存储当前样本信息
            media_info = sample["media"]
            img_path = media_info.pop("path")
            sample_item["media"] = img_path  # 添加图像样本信息
            for k, v in media_info.items():  # 添加图像相关的attributes
                this_field = Field(k, field_value=v, is_attr=True)
                sample_struct.add_field(this_field)
                sample_item[k] = v
            all_gt_info = {}

            if "ground_truth" in sample:
                gt_info = sample["ground_truth"]
            else:
                gt_info = []
                annotation_option_flag = True
                # sample_struct.set_optional("annotations")
            annotations = []
            for gt in gt_info:
                # 获取ann_id
                ann_id = gt.pop("ann_id")
                ann_id = gt.pop("ref_ann_id", ann_id)
                gt["ann_id"] = ann_id
                gt_item = {}
                gt_type = gt.pop("type")
                if gt_type == self.DETECTION_ANN_TYPE:
                    bbox_field = Field("bbox", field_type="bbox")
                    object_struct.add_field(bbox_field)
                    gt_item["bbox"] = gt.pop("bbox")
                elif gt_type == self.INSTANCE_SEG_ANN_TYPE:
                    ins_field = Field("points", field_type="points")
                    object_struct.add_field(ins_field)
                    gt_item["points"] = gt.pop("points")
                else:
                    continue
                # 创建LabelField并将值装进gt_item里面
                if 'categories' in gt and len(gt['categories']) > 0:
                    label_field = Field("category", field_type="category", param=f"${object_struct.ARG}")
                    object_struct.add_field(label_field)
                    category_id = gt.pop("categories")[0]["category_id"]
                    category_name = self.raw_class_info[gt_type][category_id]  # 拿到category_name
                    gt_item["category"] = self._class_domain.index(category_name)

                attributes = gt.pop("attributes", {})
                for attr_k, attr_v in attributes.items():
                    attr_k = Util.clean(attr_k)
                    attr_field = Field(attr_k, field_value=attr_v, is_attr=True)
                    object_struct.add_field(attr_field, optional=True)
                    gt_item[attr_k] = attr_v
                for other_k, other_v in gt.items():
                    other_k = Util.clean(other_k)
                    other_field = Field(other_k, field_value=other_v, is_attr=True)
                    object_struct.add_field(other_field)
                    gt_item[other_k] = other_v
                all_gt_info.setdefault(ann_id, {})
                all_gt_info[ann_id].update(gt_item)

            annotations.extend(list(all_gt_info.values()))
            if annotations:
                sample_item["annotations"] = annotations

            sample_container.add_item(sample_item, not self.separate_flag)

        if len(object_struct.fields):
            annotation_field = Field("annotations", field_type="list",
                                     param=object_struct.as_others_param(sample_struct))
            sample_struct.add_field(annotation_field)  # 添加annotation field
            if annotation_option_flag:
                sample_struct.set_optional("annotations")
            return (object_struct, sample_struct), sample_container
        else:
            return (sample_struct,), sample_container

    def _read_json(self, file, encoding="utf-8"):
        bytes_info = self.file_client.get(file)
        try:
            text = bytes_info.decode(encoding)
        except Exception as e:
            raise RuntimeError(f"Failed to read '{file}'. {e}") from None

        try:
            json_info = json.loads(text)
        except Exception as e:
            raise RuntimeError(f"Failed to load JSON file '{file}'. {e}") from None
        return json_info

    @property
    def struct_defs(self):
        """
        dsdl的struct定义字典组成的列表
        """
        return self._struct_defs

    @property
    def samples(self):
        """
        dsdl的data字段下的内容，为一个字典，有sample-type域samples字段，samples包含了样本列表
        """
        return self._samples

    @property
    def sample_info(self):
        """
        tunas v0.3的annotation的json文件的原始内容
        """
        return self._annotation_info

    @property
    def class_domain_info(self):
        """
        dsdl的class_domain的定义字典
        """
        return self._parent_class_domain, self._class_domain

    @property
    def meta_info(self):
        """
        dsdl的元信息内容
        """
        return self._meta_info

    @property
    def data_info_file(self):
        """
        返回tunas v0.3 dataset_info.json的文件路径
        """
        return self._data_info_file

    @property
    def annotation_file(self):
        """
        返回tunas v0.3 的标注文件json文件的路径
        """
        return self._annotation_file

    @property
    def dsdl_version(self):
        """
        dsdl的版本信息字典
        """
        return self._dsdl_version_info



import re


class FileIO:
    @staticmethod
    def clean(varStr):
        return re.sub('\W|^(?=\d)', '', varStr)

    @classmethod
    def get_oss_rel_path(cls, path: str) -> str:
        bucket_name = cls.get_bucket_name(path)
        return path.replace(f"oss://{bucket_name}/", "")

    @classmethod
    def is_oss_path(cls, path: str) -> bool:
        return path.startswith("oss://")

    @classmethod
    def get_bucket_name(cls, path: str) -> str:
        return path.replace("oss://", "").split("/")[0]

    @classmethod
    def is_ceph_path(cls, path: str) -> bool:
        """Whether the path is from ceph

        Arguments:
            path: A path.

         Returns:
             Whether the path is from ceph.
        """
        return path.startswith("s3://") or ":s3://" in path





from .file_client import CephBackend, PetrelBackend, HardDiskBackend, AliyunOSSBackend
from .fileio import FileIO

__all__ = [
    "FileIO",
    "CephBackend",
    "PetrelBackend",
    "HardDiskBackend",
    "AliyunOSSBackend",
]


from .classdomain import ClassDomain, Label
from .samples import DataSample
from .struct import Field, Struct
from .utils import Util, OptionEatAll, prepare_input

__all__ = [
    "ClassDomain",
    "Label",
    "DataSample",
    "Field",
    "Struct",
    "Util",
    "OptionEatAll",
    "prepare_input",
]



from typing import List, Optional


class ClassDomain:
    DEF = "class_domain"

    def __init__(self, name, parent_domain=None):
        self._parent_domain = parent_domain
        self._name = name
        self._classes = []
        self._classes_names = []

    def _sort_label(self):
        self._classes = sorted(self._classes, key=lambda _: _.name)
        self._classes_names = sorted(self._classes_names)

    def set_parent(self, parent_domain):
        self._parent_domain = parent_domain

    @property
    def parent_domain(self):
        return self._parent_domain

    @property
    def name(self):
        return self._name

    def index(self, name):
        self._sort_label()
        return self._classes_names.index(name) + 1

    def add_label(self, item):
        if item.name in self._classes_names:
            return
        self._classes.append(item)
        self._classes_names.append(item.name)

    def __contains__(self, item):
        return item.name in self._classes_names

    def __bool__(self):
        if self._classes:
            return True
        return False

    def format(self):
        domain_name = self.name
        if self.parent_domain:
            domain_name += f"[{self.parent_domain.name}]"
        content = dict()
        content["$def"] = self.DEF
        self._sort_label()
        classes = [_.format() for _ in self._classes]
        content["classes"] = classes
        return {domain_name: content}


class Label:

    def __init__(self, name, supercategories: Optional[List] = None):
        self._name = name
        self._supercategories = []
        if supercategories is not None:
            self._supercategories = supercategories

    @property
    def name(self):
        return self._name

    def add_supercategory(self, label):
        self._supercategories.append(label)

    @property
    def supercategories(self):
        return self._supercategories

    def format(self):
        if not self._supercategories:
            return self._name

        else:
            return self._name + f"[{', '.join([_.name for _ in self.supercategories])}]"



import os
from abc import ABCMeta, abstractmethod
from .fileio import FileIO
from typing import List, Any


def _get_basename(content: List[Any]) -> Any:
    return content[0].split("/")[-1]


class BaseStorageBackend(metaclass=ABCMeta):
    """Abstract class of storage backends.

    All backends need to implement two apis: ``get()`` and ``get_text()``.
    ``get()`` reads the file as a byte stream and ``get_text()`` reads the file
    as texts.
    """

    @abstractmethod
    def get(self, filepath):
        pass

    @abstractmethod
    def get_text(self, filepath):
        pass

    @abstractmethod
    def listdir(self, filepath):
        pass


class AliyunOSSBackend(BaseStorageBackend):
    """Aliyun OSS storage backend.

    Args:
        access_key_secret (str): access key of aliyun oss
        endpoint (str): endpoint key of aliyun oss
        access_key_id (str): your access key of aliyun oss
        bucket_name (str): your bucket name of aliyun oss
    """

    def __init__(self, bucket_name, access_key_id, access_key_secret, endpoint, **kwargs):
        try:
            import oss2
        except ImportError:
            raise ImportError("Please install oss2 to enable AliyunOSSBackend.")

        auth = oss2.Auth(access_key_id, access_key_secret)
        self.client = oss2.Bucket(auth, endpoint, bucket_name)

    def get(self, filepath):
        value_buf = self.client.get_object(filepath)
        return value_buf.read()

    def get_text(self, filepath):
        raise NotImplementedError

    def listdir(self, filepath):
        try:
            import oss2
        except ImportError:
            raise ImportError("Please install oss2 to enable AliyunOSSBackend.")
        if FileIO.is_oss_path(filepath):
            filepath = FileIO.get_oss_rel_path(filepath)
        return [obj.key for obj in oss2.ObjectIteratorV2(self.client, prefix=filepath)]


class CephBackend(BaseStorageBackend):
    """Ceph storage backend.

    Args:
        path_mapping (dict|None): path mapping dict from local path to Petrel
            path. When ``path_mapping={'src': 'dst'}``, ``src`` in ``filepath``
            will be replaced by ``dst``. Default: None.
    """

    def __init__(self, path_mapping=None):
        try:
            import ceph
        except ImportError:
            raise ImportError('Please install ceph to enable CephBackend.')

        self._client = ceph.S3Client()
        assert isinstance(path_mapping, dict) or path_mapping is None
        self.path_mapping = path_mapping

    def get(self, filepath):
        filepath = str(filepath)
        if self.path_mapping is not None:
            for k, v in self.path_mapping.items():
                filepath = filepath.replace(k, v)
        value = self._client.Get(filepath)
        value_buf = memoryview(value)
        return value_buf

    def get_text(self, filepath):
        raise NotImplementedError

    def listdir(self, filepath):
        raise NotImplementedError


class PetrelBackend(BaseStorageBackend):
    """Petrel storage backend (for internal use).

    Args:
        path_mapping (dict|None): path mapping dict from local path to Petrel
            path. When `path_mapping={'src': 'dst'}`, `src` in `filepath` will
            be replaced by `dst`. Default: None.
        enable_mc (bool): whether to enable memcached support. Default: True.
    """

    def __init__(self, path_mapping=None, enable_mc=True):
        try:
            from petrel_client import client
        except ImportError:
            raise ImportError('Please install petrel_client to enable '
                              'PetrelBackend.')

        self._client = client.Client(enable_mc=enable_mc)
        assert isinstance(path_mapping, dict) or path_mapping is None
        self.path_mapping = path_mapping

    def get(self, filepath):
        filepath = str(filepath)
        if self.path_mapping is not None:
            for k, v in self.path_mapping.items():
                filepath = filepath.replace(k, v)
        value = self._client.Get(filepath)
        value_buf = memoryview(value)
        return value_buf

    def get_text(self, filepath):
        raise NotImplementedError

    def listdir(self, filepath):
        return self._client.list(filepath)


class HardDiskBackend(BaseStorageBackend):
    """Raw hard disks storage backend."""

    def get(self, filepath):
        filepath = str(filepath)
        with open(filepath, 'rb') as f:
            value_buf = f.read()
        return value_buf

    def get_text(self, filepath):
        filepath = str(filepath)
        with open(filepath, 'r') as f:
            value_buf = f.read()
        return value_buf

    def listdir(self, filepath):
        return os.listdir(filepath)



from .utils import Util


class Field:
    FIELD_MAPPING = {
        "bbox": "BBox",
        "int": "Int",
        "float": "Num",
        "str": "Str",
        "points": "Polygon",
        "list": "List",
        "media": "Image",
        "bool": "Bool",
        "category": f"Label",
    }

    def __init__(self, name, field_value=None, field_type=None, is_attr=None, param=None):
        self._name = name
        self._is_attr = is_attr
        self.param = param
        self.arg = None
        if field_type is not None:
            self.field_type = self.FIELD_MAPPING[field_type]
        elif field_value is not None:
            self.field_type = self.FIELD_MAPPING[field_value.__class__.__name__]
            if self.field_type == "List":
                self.arg = "etype"
                self.param = self.FIELD_MAPPING[field_value[0].__class__.__name__]
        else:
            raise RuntimeError("not supported field type")
        if field_type == "category":
            self.arg = "dom"
            assert isinstance(self.param, str)
        elif field_type == "list":
            self.arg = "etype"
            assert isinstance(self.param, str)

    @property
    def is_attr(self):
        return self._is_attr

    @property
    def name(self):
        return self._name

    def format(self):
        if not self._is_attr and not self.arg:
            return {self._name: self.field_type}
        elif self.is_attr and not self.arg:
            return {self._name: f"{self.field_type}[is_attr=True]"}
        elif not self.is_attr and self.arg:
            return {self._name: f"{self.field_type}[{self.arg}={self.param}]"}
        else:
            return {self._name: f"{self.field_type}[is_attr=True, {self.arg}={self.param}]"}


class Struct:
    DEF = "struct"
    ARG = "cdom"

    def __init__(self, name, arg=None):
        self._name = name
        self.fields = []
        self.field_names = []
        self._optional = []
        if arg is not None:
            self.ARG = arg

    def add_field(self, item, optional=False):
        if item.name in self.field_names:
            return
        self.fields.append(item)
        self.field_names.append(item.name)
        if optional:
            self._optional.append(item.name)

    def set_optional(self, name):
        if name in self.field_names:
            self._optional.append(name)
            return
        else:
            raise RuntimeError(f"field {name} not exists.")

    @property
    def name(self):
        return self._name

    def format(self):
        content = dict()
        content["$def"] = self.DEF
        content["$params"] = Util.flist([self.ARG])
        content["$fields"] = {}
        for field in self.fields:
            content["$fields"].update(field.format())
        optional_lst = list(set(self._optional))
        if optional_lst:
            content["$optional"] = Util.flist(optional_lst)
        return {self.name: content}

    def as_others_param(self, other_struct):
        return f"{self._name}[{self.ARG}=${other_struct.ARG}]"



from .utils import Util


class DataSample:
    def __init__(self, struct, class_domain):
        self.sample_struct = struct
        self.class_domain = class_domain
        self.samples = []

    def add_item(self, item, pre_format=True):
        if pre_format:
            sample = self.add_quotes(item.copy())
            for k, v in sample.items():
                if Util.is_list_of(v, dict):
                    sample[k] = [Util.fmap(_) for _ in v]
            self.samples.append(sample)
        else:
            self.samples.append(item)

    def add_quotes(self, sample):
        if isinstance(sample, dict):
            for k, v in sample.items():
                sample[k] = self.add_quotes(v)
        elif isinstance(sample, list):
            for k, v in enumerate(sample):
                sample[k] = self.add_quotes(v)
        elif isinstance(sample, str):
            sample = Util.add_quotes(sample)

        return sample

    @property
    def sample_type(self):
        return f"{self.sample_struct.name}[{self.sample_struct.ARG}={self.class_domain.name}]"

    def format(self, file_name="local"):
        content = dict()
        content["sample-type"] = self.sample_type
        content["sample-path"] = file_name
        content["samples"] = self.samples
        return {"data": content}



import click

try:
    import ruamel_yaml as yaml
    from ruamel_yaml import YAML
except ImportError:
    import ruamel.yaml as yaml
    from ruamel.yaml import YAML
from ..io import FileIO, PetrelBackend, AliyunOSSBackend, HardDiskBackend
import os


class Util:

    @staticmethod
    def is_list_of(obj, cls):
        if isinstance(obj, list):
            return isinstance(obj[0], cls)
        return False

    @staticmethod
    def add_quotes(string):
        # string = f'"{string}"'
        string = yaml.scalarstring.DoubleQuotedScalarString(string)
        return string

    @staticmethod
    def flist(x):
        x = [Util.add_quotes(_) for _ in x]
        retval = yaml.comments.CommentedSeq(x)
        retval.fa.set_flow_style()  # fa -> format attribute
        return retval

    @staticmethod
    def fmap(x):
        retval = yaml.comments.CommentedMap(x)
        retval.fa.set_flow_style()  # fa -> format attribute
        return retval

    @staticmethod
    def clean(s: str):
        s = s.replace('/', '')
        s = s.strip()
        if s.isidentifier():
            s = "_" + s
        return s


class OptionEatAll(click.Option):
    """
    implemented by https://stackoverflow.com/users/7311767/stephen-rauch in https://stackoverflow.com/questions/48391777/nargs-equivalent-for-options-in-click
    """

    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):

        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


def prepare_input(func):
    def process(directory, dataset_info, annotation_info, working_dir, task, config, separate_store_flag):

        assert os.path.isdir(working_dir), f"The working dir '{working_dir}' is not a directory."

        if config:
            import json
            with open(config, "r") as f:
                config = json.load(f)

        if directory is not None:
            assert not (
                    dataset_info or annotation_info), "You have specified the directory, which means you don't need to specify the paths of annotation file and `dataset_info.json`."
            if FileIO.is_ceph_path(directory):
                file_client = PetrelBackend()
                dataset_info = os.path.join(directory, 'dataset_info.json')
                ann_dir = os.path.join(directory, 'annotations', 'json', )
                annotation_info = [os.path.join(ann_dir, _) for _ in file_client.listdir(ann_dir) if
                                   _.endswith('.json')]
            elif FileIO.is_oss_path(directory):
                assert config, "you must provide the configuration of aliyun oss."
                bucket_name = FileIO.get_bucket_name(directory)
                file_client = AliyunOSSBackend(bucket_name=bucket_name, **config)
                dataset_info = os.path.join(directory, 'dataset_info.json')
                dataset_info = FileIO.get_oss_rel_path(dataset_info)
                if directory.endswith("/"):
                    directory = directory[:-1]
                ann_dir = '/'.join([directory, 'annotations', 'json'])
                annotation_info = [_ for _ in file_client.listdir(ann_dir) if _.endswith('.json')]
            else:
                file_client = HardDiskBackend()
                dataset_info = os.path.join(directory, 'dataset_info.json')
                ann_dir = os.path.join(directory, 'annotations', 'json')
                annotation_info = [os.path.join(ann_dir, _) for _ in file_client.listdir(ann_dir) if
                                   _.endswith('.json')]

        else:
            assert dataset_info and annotation_info, "The directory of dataset is not speicified, which means you need to specify the paths of annotation file and `dataset_info.json` with '-a' and '-i'"
            if FileIO.is_ceph_path(dataset_info):
                file_client = PetrelBackend()
                annotation_info = [_ for _ in annotation_info if _.endswith('.json')]
            elif FileIO.is_oss_path(dataset_info):
                assert config, "you must provide the configuration of aliyun oss."
                bucket_name = FileIO.get_bucket_name(dataset_info)
                bucket_name_ = FileIO.get_bucket_name(annotation_info)
                assert bucket_name_ == bucket_name, "the files must be from the same bucket."
                file_client = AliyunOSSBackend(bucket_name=bucket_name, **config)
                dataset_info = FileIO.get_oss_rel_path(dataset_info)
                annotation_info = [FileIO.get_oss_rel_path(_) for _ in annotation_info if _.endswith('.json')]
            else:
                file_client = HardDiskBackend()
                annotation_info = [_ for _ in annotation_info if _.endswith('.json')]

        annotation_info = {os.path.basename(_).split('.')[0]: _ for _ in annotation_info}
        return directory, dataset_info, annotation_info, working_dir, task, file_client, separate_store_flag

    def wrapper(*args, **kwargs):
        ret = process(*args, **kwargs)
        ret = func(*ret)
        return ret

    return wrapper


def get_segment_name(filepath):
    segment = os.path.basename(filepath).split('.')[0]
    return segment
